@@ -1,5 +1,6 @@
 /* keyserver.c - generic keyserver code
- * Copyright (C) 2001, 2002, 2003, 2004, 2005 Free Software Foundation, Inc.
+ * Copyright (C) 2001, 2002, 2003, 2004, 2005,
+ *               2006 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -43,12 +44,18 @@
 #include "keyserver-internal.h"
 #include "util.h"
 
-#define GET    0
-#define SEND   1
-#define SEARCH 2
-
 #define GPGKEYS_PREFIX "gpgkeys_"
 
+#if defined(HAVE_LIBCURL) || defined(FAKE_CURL)
+#define GPGKEYS_CURL "gpgkeys_curl"
+#endif
+
+#ifdef GPGKEYS_CURL
+#define GPGKEYS_PREFIX_LEN (strlen(GPGKEYS_PREFIX)+strlen(GPGKEYS_CURL))
+#else
+#define GPGKEYS_PREFIX_LEN (strlen(GPGKEYS_PREFIX))
+#endif
+
 struct keyrec
 {
   KEYDB_SEARCH_DESC desc;
@@ -59,29 +66,69 @@ struct keyrec
   unsigned int lines;
 };
 
-/* Tell remote processes about these options */
-#define REMOTE_TELL (KEYSERVER_INCLUDE_REVOKED|KEYSERVER_INCLUDE_SUBKEYS|KEYSERVER_TRY_DNS_SRV)
+enum ks_action {KS_UNKNOWN=0,KS_GET,KS_GETNAME,KS_SEND,KS_SEARCH};
 
 static struct parse_options keyserver_opts[]=
   {
-    {"include-revoked",KEYSERVER_INCLUDE_REVOKED,NULL},      
-    {"include-subkeys",KEYSERVER_INCLUDE_SUBKEYS,NULL},
-    {"keep-temp-files",KEYSERVER_KEEP_TEMP_FILES,NULL},
-    {"refresh-add-fake-v3-keyids",KEYSERVER_ADD_FAKE_V3,NULL},
-    {"auto-key-retrieve",KEYSERVER_AUTO_KEY_RETRIEVE,NULL},
-    {"try-dns-srv",KEYSERVER_TRY_DNS_SRV,NULL},
-    {"honor-keyserver-url",KEYSERVER_HONOR_KEYSERVER_URL,NULL},
-    {NULL,0,NULL}
+    /* some of these options are not real - just for the help
+       message */
+    {"max-cert-size",0,NULL,NULL},
+    {"include-revoked",0,NULL,N_("include revoked keys in search results")},
+    {"include-subkeys",0,NULL,N_("include subkeys when searching by key ID")},
+    {"use-temp-files",0,NULL,
+     N_("use temporary files to pass data to keyserver helpers")},
+    {"keep-temp-files",KEYSERVER_KEEP_TEMP_FILES,NULL,
+     N_("do not delete temporary files after using them")},
+    {"refresh-add-fake-v3-keyids",KEYSERVER_ADD_FAKE_V3,NULL,
+     NULL},
+    {"auto-key-retrieve",KEYSERVER_AUTO_KEY_RETRIEVE,NULL,
+     N_("automatically retrieve keys when verifying signatures")},
+    {"honor-keyserver-url",KEYSERVER_HONOR_KEYSERVER_URL,NULL,
+     N_("honor the preferred keyserver URL set on the key")},
+    {"honor-pka-record",KEYSERVER_HONOR_PKA_RECORD,NULL,
+     N_("honor the PKA record set on a key when retrieving keys")},
+    {NULL,0,NULL,NULL}
   };
 
-static int keyserver_work(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
-			  int count,struct keyserver_spec *keyserver);
+static int keyserver_work(enum ks_action action,STRLIST list,
+			  KEYDB_SEARCH_DESC *desc,int count,
+			  unsigned char **fpr,size_t *fpr_len,
+			  struct keyserver_spec *keyserver);
+
+/* Reasonable guess */
+#define DEFAULT_MAX_CERT_SIZE 16384
+
+static size_t max_cert_size=DEFAULT_MAX_CERT_SIZE;
+
+static void
+add_canonical_option(char *option,STRLIST *list)
+{
+  char *arg=argsplit(option);
+
+  if(arg)
+    {
+      char *joined;
+
+      joined=xmalloc(strlen(option)+1+strlen(arg)+1);
+      /* Make a canonical name=value form with no spaces */
+      strcpy(joined,option);
+      strcat(joined,"=");
+      strcat(joined,arg);
+      append_to_strlist(list,joined);
+      xfree(joined);
+    }
+  else
+    append_to_strlist(list,option);
+}
 
 int
 parse_keyserver_options(char *options)
 {
   int ret=1;
   char *tok;
+  char *max_cert=NULL;
+
+  keyserver_opts[0].value=&max_cert;
 
   while((tok=optsep(&options)))
     {
@@ -122,40 +169,68 @@ parse_keyserver_options(char *options)
 	{
 	  /* All of the standard options have failed, so the option is
 	     destined for a keyserver plugin. */
-	  char *arg=argsplit(tok);
-
-	  if(arg)
-	    {
-	      char *joined;
-
-	      joined=m_alloc(strlen(tok)+1+strlen(arg)+1);
-	      /* Make a canonical name=value form with no
-		 spaces */
-	      strcpy(joined,tok);
-	      strcat(joined,"=");
-	      strcat(joined,arg);
-	      add_to_strlist(&opt.keyserver_options.other,joined);
-	      m_free(joined);
-	    }
-	  else
-	    add_to_strlist(&opt.keyserver_options.other,tok);
+	  add_canonical_option(tok,&opt.keyserver_options.other);
 	}
     }
 
+  if(max_cert)
+    {
+      max_cert_size=strtoul(max_cert,(char **)NULL,10);
+
+      if(max_cert_size==0)
+	max_cert_size=DEFAULT_MAX_CERT_SIZE;
+    }
+
   return ret;
 }
 
 void
 free_keyserver_spec(struct keyserver_spec *keyserver)
 {
-  m_free(keyserver->uri);
-  m_free(keyserver->scheme);
-  m_free(keyserver->auth);
-  m_free(keyserver->host);
-  m_free(keyserver->port);
-  m_free(keyserver->path);
-  m_free(keyserver->opaque);
-  m_free(keyserver);
+  xfree(keyserver->uri);
+  xfree(keyserver->scheme);
+  xfree(keyserver->auth);
+  xfree(keyserver->host);
+  xfree(keyserver->port);
+  xfree(keyserver->path);
+  xfree(keyserver->opaque);
+  free_strlist(keyserver->options);
+  xfree(keyserver);
+}
+
+/* Return 0 for match */
+static int
+cmp_keyserver_spec(struct keyserver_spec *one,struct keyserver_spec *two)
+{
+  if(ascii_strcasecmp(one->scheme,two->scheme)==0)
+    {
+      if(one->host && two->host && ascii_strcasecmp(one->host,two->host)==0)
+	{
+	  if((one->port && two->port
+	      && ascii_strcasecmp(one->port,two->port)==0)
+	     || (!one->port && !two->port))
+	    return 0;
+	}
+      else if(one->opaque && two->opaque
+	      && ascii_strcasecmp(one->opaque,two->opaque)==0)
+	return 0;
+    }
+
+  return 1;
+}
+
+/* Try and match one of our keyservers.  If we can, return that.  If
+   we can't, return our input. */
+struct keyserver_spec *
+keyserver_match(struct keyserver_spec *spec)
+{
+  struct keyserver_spec *ks;
+
+  for(ks=opt.keyserver;ks;ks=ks->next)
+    if(cmp_keyserver_spec(spec,ks)==0)
+      return ks;
+
+  return spec;
 }
 
 /* TODO: once we cut over to an all-curl world, we don't need this
@@ -163,19 +238,32 @@ free_keyserver_spec(struct keyserver_spec *keyserver)
    keyserver/ksutil.c for limited use in gpgkeys_ldap or the like. */
 
 struct keyserver_spec *
-parse_keyserver_uri(const char *uri,int require_scheme,
+parse_keyserver_uri(const char *string,int require_scheme,
 		    const char *configname,unsigned int configlineno)
 {
   int assume_hkp=0;
   struct keyserver_spec *keyserver;
   const char *idx;
   int count;
+  char *uri,*options;
 
-  assert(uri!=NULL);
+  assert(string!=NULL);
 
-  keyserver=m_alloc_clear(sizeof(struct keyserver_spec));
+  keyserver=xmalloc_clear(sizeof(struct keyserver_spec));
 
-  keyserver->uri=m_strdup(uri);
+  uri=xstrdup(string);
+
+  options=strchr(uri,' ');
+  if(options)
+    {
+      char *tok;
+
+      *options='\0';
+      options++;
+
+      while((tok=optsep(&options)))
+	add_canonical_option(tok,&keyserver->options);
+    }
 
   /* Get the scheme */
 
@@ -206,13 +294,20 @@ parse_keyserver_uri(const char *uri,int require_scheme,
 
       /* Assume HKP if there is no scheme */
       assume_hkp=1;
-      keyserver->scheme=m_strdup("hkp");
+      keyserver->scheme=xstrdup("hkp");
+
+      keyserver->uri=xmalloc(strlen(keyserver->scheme)+3+strlen(uri)+1);
+      strcpy(keyserver->uri,keyserver->scheme);
+      strcat(keyserver->uri,"://");
+      strcat(keyserver->uri,uri);
     }
   else
     {
       int i;
 
-      keyserver->scheme=m_alloc(count+1);
+      keyserver->uri=xstrdup(uri);
+
+      keyserver->scheme=xmalloc(count+1);
 
       /* Force to lowercase */
       for(i=0;i<count;i++)
@@ -228,16 +323,16 @@ parse_keyserver_uri(const char *uri,int require_scheme,
     {
       deprecated_warning(configname,configlineno,"x-broken-hkp",
 			 "--keyserver-options ","broken-http-proxy");
-      m_free(keyserver->scheme);
-      keyserver->scheme=m_strdup("hkp");
-      add_to_strlist(&opt.keyserver_options.other,"broken-http-proxy");
+      xfree(keyserver->scheme);
+      keyserver->scheme=xstrdup("hkp");
+      append_to_strlist(&opt.keyserver_options.other,"broken-http-proxy");
     }
   else if(ascii_strcasecmp(keyserver->scheme,"x-hkp")==0)
     {
       /* Canonicalize this to "hkp" so it works with both the internal
 	 and external keyserver interface. */
-      m_free(keyserver->scheme);
-      keyserver->scheme=m_strdup("hkp");
+      xfree(keyserver->scheme);
+      keyserver->scheme=xstrdup("hkp");
     }
 
   if(assume_hkp || (uri[0]=='/' && uri[1]=='/'))
@@ -259,7 +354,7 @@ parse_keyserver_uri(const char *uri,int require_scheme,
 	  if(count==0)
 	    goto fail;
 
-	  keyserver->auth=m_alloc(count+1);
+	  keyserver->auth=xmalloc(count+1);
 	  strncpy(keyserver->auth,uri,count);
 	  keyserver->auth[count]='\0';
 	  uri+=count+1;
@@ -286,7 +381,7 @@ parse_keyserver_uri(const char *uri,int require_scheme,
       if(count==0)
 	goto fail;
 
-      keyserver->host=m_alloc(count+1);
+      keyserver->host=xmalloc(count+1);
       strncpy(keyserver->host,uri,count);
       keyserver->host[count]='\0';
 
@@ -309,7 +404,7 @@ parse_keyserver_uri(const char *uri,int require_scheme,
 		goto fail;
 	    }
 
-	  keyserver->port=m_alloc(count+1);
+	  keyserver->port=xmalloc(count+1);
 	  strncpy(keyserver->port,uri+1,count);
 	  keyserver->port[count]='\0';
 
@@ -319,15 +414,18 @@ parse_keyserver_uri(const char *uri,int require_scheme,
 
       /* Everything else is the path */
       if(*uri)
-	keyserver->path=m_strdup(uri);
+	keyserver->path=xstrdup(uri);
       else
-	keyserver->path=m_strdup("/");
+	keyserver->path=xstrdup("/");
+
+      if(keyserver->path[1]!='\0')
+	keyserver->flags.direct_uri=1;
     }
   else if(uri[0]!='/')
     {
       /* No slash means opaque.  Just record the opaque blob and get
 	 out. */
-      keyserver->opaque=m_strdup(uri);
+      keyserver->opaque=xstrdup(uri);
     }
   else
     {
@@ -354,12 +452,12 @@ parse_preferred_keyserver(PKT_signature *sig)
   p=parse_sig_subpkt(sig->hashed,SIGSUBPKT_PREF_KS,&plen);
   if(p && plen)
     {
-      byte *dupe=m_alloc(plen+1);
+      byte *dupe=xmalloc(plen+1);
 
       memcpy(dupe,p,plen);
       dupe[plen]='\0';
       spec=parse_keyserver_uri(dupe,1,NULL,0);
-      m_free(dupe);
+      xfree(dupe);
     }
 
   return spec;
@@ -461,7 +559,7 @@ parse_keyrec(char *keystring)
 	return NULL;
       else if(work->desc.mode==KEYDB_SEARCH_MODE_NONE)
 	{
-	  m_free(work);
+	  xfree(work);
 	  return NULL;
 	}
       else
@@ -474,7 +572,7 @@ parse_keyrec(char *keystring)
 
   if(work==NULL)
     {
-      work=m_alloc_clear(sizeof(struct keyrec));
+      work=xmalloc_clear(sizeof(struct keyrec));
       work->uidbuf=iobuf_temp();
     }
 
@@ -495,7 +593,7 @@ parse_keyrec(char *keystring)
       if(work->desc.mode)
 	{
 	  ret=work;
-	  work=m_alloc_clear(sizeof(struct keyrec));
+	  work=xmalloc_clear(sizeof(struct keyrec));
 	  work->uidbuf=iobuf_temp();
 	}
 
@@ -610,7 +708,7 @@ parse_keyrec(char *keystring)
       if(strlen(decoded)>opt.screen_columns-10)
 	decoded[opt.screen_columns-10]='\0';
       iobuf_writestr(work->uidbuf,decoded);
-      m_free(decoded);
+      xfree(decoded);
       iobuf_writestr(work->uidbuf,"\n\t");
       work->lines++;
     }
@@ -648,7 +746,7 @@ show_prompt(KEYDB_SEARCH_DESC *desc,int numdesc,int count,const char *search)
 
   if(answer[0]=='q' || answer[0]=='Q')
     {
-      m_free(answer);
+      xfree(answer);
       return 1;
     }
   else if(atoi(answer)>=1 && atoi(answer)<=numdesc)
@@ -657,9 +755,10 @@ show_prompt(KEYDB_SEARCH_DESC *desc,int numdesc,int count,const char *search)
 
       while((num=strsep(&split," ,"))!=NULL)
 	if(atoi(num)>=1 && atoi(num)<=numdesc)
-	  keyserver_work(GET,NULL,&desc[atoi(num)-1],1,opt.keyserver);
+	  keyserver_work(KS_GET,NULL,&desc[atoi(num)-1],1,
+			 NULL,NULL,opt.keyserver);
 
-      m_free(answer);
+      xfree(answer);
       return 1;
     }
 
@@ -681,7 +780,7 @@ keyserver_search_prompt(IOBUF buffer,const char *searchstr)
   if(searchstr)
     localstr=utf8_to_native(searchstr,strlen(searchstr),0);
 
-  desc=m_alloc(count*sizeof(KEYDB_SEARCH_DESC));
+  desc=xmalloc(count*sizeof(KEYDB_SEARCH_DESC));
 
   for(;;)
     {
@@ -736,7 +835,7 @@ keyserver_search_prompt(IOBUF buffer,const char *searchstr)
 	      else
 		validcount=1;
 
-	      desc=m_realloc(desc,count*sizeof(KEYDB_SEARCH_DESC));
+	      desc=xrealloc(desc,count*sizeof(KEYDB_SEARCH_DESC));
 	    }
 
 	  started=1;
@@ -776,7 +875,7 @@ keyserver_search_prompt(IOBUF buffer,const char *searchstr)
 	  /* keyserver helper sent more keys than they claimed in the
 	     info: line. */
 	  count+=10;
-	  desc=m_realloc(desc,count*sizeof(KEYDB_SEARCH_DESC));
+	  desc=xrealloc(desc,count*sizeof(KEYDB_SEARCH_DESC));
 	  validcount=0;
 	}
 
@@ -800,7 +899,7 @@ keyserver_search_prompt(IOBUF buffer,const char *searchstr)
 
 	  numlines+=keyrec->lines;
 	  iobuf_close(keyrec->uidbuf);
-	  m_free(keyrec);
+	  xfree(keyrec);
 
 	  started=1;
 	  i++;
@@ -824,34 +923,9 @@ keyserver_search_prompt(IOBUF buffer,const char *searchstr)
 	log_info(_("key not found on keyserver\n"));
     }
 
-  m_free(localstr);
-  m_free(desc);
-  m_free(line);
-}
-
-static int
-curl_can_handle(const char *scheme)
-{
-#if defined(HAVE_LIBCURL)
-
-  const char * const *proto;
-  curl_version_info_data *data=curl_version_info(CURLVERSION_NOW);
-
-  assert(data);
-
-  for(proto=data->protocols;*proto;proto++)
-    if(strcasecmp(*proto,scheme)==0)
-      return 1;
-
-#elif defined(FAKE_CURL)
-
-  /* If we're faking curl, then we only support HTTP */
-  if(strcasecmp(scheme,"http")==0)
-    return 1;
-
-#endif
-
-  return 0;
+  xfree(localstr);
+  xfree(desc);
+  xfree(line);
 }
 
 /* We sometimes want to use a different gpgkeys_xxx for a given
@@ -862,25 +936,36 @@ keyserver_typemap(const char *type)
 {
   if(strcmp(type,"ldaps")==0)
     return "ldap";
-  else if(curl_can_handle(type))
-    return "curl";
   else
     return type;
 }
 
+#ifdef GPGKEYS_CURL
+/* The PGP LDAP and the curl fetch-a-LDAP-object methodologies are
+   sufficiently different that we can't use curl to do LDAP. */
+static int
+curl_cant_handle(const char *scheme,unsigned int direct_uri)
+{
+  if(!direct_uri && (strcmp(scheme,"ldap")==0 || strcmp(scheme,"ldaps")==0))
+    return 1;
+
+  return 0;
+}
+#endif
+
 #define KEYSERVER_ARGS_KEEP " -o \"%O\" \"%I\""
 #define KEYSERVER_ARGS_NOKEEP " -o \"%o\" \"%i\""
 
 static int 
-keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
-		int count,int *prog,struct keyserver_spec *keyserver)
+keyserver_spawn(enum ks_action action,STRLIST list,KEYDB_SEARCH_DESC *desc,
+		int count,int *prog,unsigned char **fpr,size_t *fpr_len,
+		struct keyserver_spec *keyserver)
 {
   int ret=0,i,gotversion=0,outofband=0;
   STRLIST temp;
   unsigned int maxlen,buflen;
-  char *command,*searchstr=NULL;
+  char *command,*end,*searchstr=NULL;
   byte *line=NULL;
-  struct parse_options *kopts;
   struct exec_info *spawn;
   const char *scheme;
   const char *libexecdir = get_libexecdir ();
@@ -920,33 +1005,46 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
       /* If exec-path was set, and DISABLE_KEYSERVER_PATH is
 	 undefined, then don't specify a full path to gpgkeys_foo, so
 	 that the PATH can work. */
-      command=m_alloc(strlen(GPGKEYS_PREFIX)+strlen(scheme)+1);
+      command=xmalloc(GPGKEYS_PREFIX_LEN+strlen(scheme)+3+strlen(EXEEXT)+1);
       command[0]='\0';
     }
   else
 #endif
     {
       /* Specify a full path to gpgkeys_foo. */
-      command=m_alloc(strlen(libexecdir)+strlen(DIRSEP_S)+
-		      strlen(GPGKEYS_PREFIX)+strlen(scheme)+1);
+      command=xmalloc(strlen(libexecdir)+strlen(DIRSEP_S)+
+		      GPGKEYS_PREFIX_LEN+strlen(scheme)+3+strlen(EXEEXT)+1);
       strcpy(command,libexecdir);
       strcat(command,DIRSEP_S);
     }
 
+  end=command+strlen(command);
+
   strcat(command,GPGKEYS_PREFIX); 
   strcat(command,scheme);
 
+  if(keyserver->flags.direct_uri)
+    strcat(command,"uri");
+
+  strcat(command,EXEEXT);
+
+#ifdef GPGKEYS_CURL
+  if(!curl_cant_handle(scheme,keyserver->flags.direct_uri)
+     && path_access(command,X_OK)!=0)
+    strcpy(end,GPGKEYS_CURL);
+#endif
+
   if(opt.keyserver_options.options&KEYSERVER_USE_TEMP_FILES)
     {
       if(opt.keyserver_options.options&KEYSERVER_KEEP_TEMP_FILES)
 	{
-	  command=m_realloc(command,strlen(command)+
+	  command=xrealloc(command,strlen(command)+
 			    strlen(KEYSERVER_ARGS_KEEP)+1);
 	  strcat(command,KEYSERVER_ARGS_KEEP);
 	}
       else
 	{
-	  command=m_realloc(command,strlen(command)+
+	  command=xrealloc(command,strlen(command)+
 			    strlen(KEYSERVER_ARGS_NOKEEP)+1);
 	  strcat(command,KEYSERVER_ARGS_NOKEEP);  
 	}
@@ -956,7 +1054,7 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
   else
     ret=exec_write(&spawn,command,NULL,NULL,0,0);
 
-  m_free(command);
+  xfree(command);
 
   if(ret)
     return ret;
@@ -984,18 +1082,19 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	fprintf(spawn->tochild,"PATH %s\n",keyserver->path);
     }
 
-  /* Write options */
-
-  for(i=0,kopts=keyserver_opts;kopts[i].name;i++)
-    if(opt.keyserver_options.options & kopts[i].bit & REMOTE_TELL)
-      fprintf(spawn->tochild,"OPTION %s\n",kopts[i].name);
+  /* Write global options */
 
   for(temp=opt.keyserver_options.other;temp;temp=temp->next)
     fprintf(spawn->tochild,"OPTION %s\n",temp->d);
 
+  /* Write per-keyserver options */
+
+  for(temp=keyserver->options;temp;temp=temp->next)
+    fprintf(spawn->tochild,"OPTION %s\n",temp->d);
+
   switch(action)
     {
-    case GET:
+    case KS_GET:
       {
 	fprintf(spawn->tochild,"COMMAND GET\n\n");
 
@@ -1003,6 +1102,8 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 
 	for(i=0;i<count;i++)
 	  {
+	    int quiet=0;
+
 	    if(desc[i].mode==KEYDB_SEARCH_MODE_FPR20)
 	      {
 		int f;
@@ -1032,18 +1133,26 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	    else if(desc[i].mode==KEYDB_SEARCH_MODE_SHORT_KID)
 	      fprintf(spawn->tochild,"0x%08lX\n",
 		      (ulong)desc[i].u.kid[1]);
+	    else if(desc[i].mode==KEYDB_SEARCH_MODE_EXACT)
+	      {
+		fprintf(spawn->tochild,"0x0000000000000000\n");
+		quiet=1;
+	      }
 	    else if(desc[i].mode==KEYDB_SEARCH_MODE_NONE)
 	      continue;
 	    else
 	      BUG();
 
-	    if(keyserver->host)
-	      log_info(_("requesting key %s from %s server %s\n"),
-		       keystr_from_desc(&desc[i]),
-		       keyserver->scheme,keyserver->host);
-	    else
-	      log_info(_("requesting key %s from %s\n"),
-		       keystr_from_desc(&desc[i]),keyserver->uri);
+	    if(!quiet)
+	      {
+		if(keyserver->host)
+		  log_info(_("requesting key %s from %s server %s\n"),
+			   keystr_from_desc(&desc[i]),
+			   keyserver->scheme,keyserver->host);
+		else
+		  log_info(_("requesting key %s from %s\n"),
+			   keystr_from_desc(&desc[i]),keyserver->uri);
+	      }
 	  }
 
 	fprintf(spawn->tochild,"\n");
@@ -1051,7 +1160,29 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	break;
       }
 
-    case SEND:
+    case KS_GETNAME:
+      {
+	STRLIST key;
+
+	fprintf(spawn->tochild,"COMMAND GETNAME\n\n");
+
+	/* Which names do we want? */
+
+	for(key=list;key!=NULL;key=key->next)
+	  fprintf(spawn->tochild,"%s\n",key->d);
+
+	fprintf(spawn->tochild,"\n");
+
+	if(keyserver->host)
+	  log_info(_("searching for names from %s server %s\n"),
+		   keyserver->scheme,keyserver->host);
+	else
+	  log_info(_("searching for names from %s\n"),keyserver->uri);
+
+	break;
+      }
+
+    case KS_SEND:
       {
 	STRLIST key;
 
@@ -1211,7 +1342,7 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	break;
       }
 
-    case SEARCH:
+    case KS_SEARCH:
       {
 	STRLIST key;
 
@@ -1225,13 +1356,13 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	    fprintf(spawn->tochild,"%s\n",key->d);
 	    if(key!=list)
 	      {
-		searchstr=m_realloc(searchstr,
+		searchstr=xrealloc(searchstr,
 				    strlen(searchstr)+strlen(key->d)+2);
 		strcat(searchstr," ");
 	      }
 	    else
 	      {
-		searchstr=m_alloc(strlen(key->d)+1);
+		searchstr=xmalloc(strlen(key->d)+1);
 		searchstr[0]='\0';
 	      }
 
@@ -1315,7 +1446,8 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
   if(!outofband)
     switch(action)
       {
-      case GET:
+      case KS_GET:
+      case KS_GETNAME:
 	{
 	  void *stats_handle;
 
@@ -1328,7 +1460,7 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	     way to do this could be to continue parsing this
 	     line-by-line and make a temp iobuf for each key. */
 
-	  import_keys_stream(spawn->fromchild,stats_handle,
+	  import_keys_stream(spawn->fromchild,stats_handle,fpr,fpr_len,
 			     opt.keyserver_options.import_options);
 
 	  import_print_stats(stats_handle);
@@ -1338,10 +1470,10 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	}
 
 	/* Nothing to do here */
-      case SEND:
+      case KS_SEND:
 	break;
 
-      case SEARCH:
+      case KS_SEARCH:
 	keyserver_search_prompt(spawn->fromchild,searchstr);
 	break;
 
@@ -1361,8 +1493,9 @@ keyserver_spawn(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 }
 
 static int 
-keyserver_work(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
-	       int count,struct keyserver_spec *keyserver)
+keyserver_work(enum ks_action action,STRLIST list,KEYDB_SEARCH_DESC *desc,
+	       int count,unsigned char **fpr,size_t *fpr_len,
+	       struct keyserver_spec *keyserver)
 {
   int rc=0,ret=0;
 
@@ -1380,7 +1513,7 @@ keyserver_work(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 #else
   /* Spawn a handler */
 
-  rc=keyserver_spawn(action,list,desc,count,&ret,keyserver);
+  rc=keyserver_spawn(action,list,desc,count,&ret,fpr,fpr_len,keyserver);
   if(ret)
     {
       switch(ret)
@@ -1393,8 +1526,8 @@ keyserver_work(int action,STRLIST list,KEYDB_SEARCH_DESC *desc,
 	case KEYSERVER_NOT_SUPPORTED:
 	  log_error(_("action `%s' not supported with keyserver "
 		      "scheme `%s'\n"),
-		    action==GET?"get":action==SEND?"send":
-		    action==SEARCH?"search":"unknown",
+		    action==KS_GET?"get":action==KS_SEND?"send":
+		    action==KS_SEARCH?"search":"unknown",
 		    keyserver->scheme);
 	  break;
 
@@ -1454,7 +1587,7 @@ keyserver_export(STRLIST users)
 
   if(sl)
     {
-      rc=keyserver_work(SEND,sl,NULL,0,opt.keyserver);
+      rc=keyserver_work(KS_SEND,sl,NULL,0,NULL,NULL,opt.keyserver);
       free_strlist(sl);
     }
 
@@ -1469,7 +1602,7 @@ keyserver_import(STRLIST users)
   int rc=0;
 
   /* Build a list of key ids */
-  desc=m_alloc(sizeof(KEYDB_SEARCH_DESC)*num);
+  desc=xmalloc(sizeof(KEYDB_SEARCH_DESC)*num);
 
   for(;users;users=users->next)
     {
@@ -1487,14 +1620,14 @@ keyserver_import(STRLIST users)
       if(count==num)
 	{
 	  num+=100;
-	  desc=m_realloc(desc,sizeof(KEYDB_SEARCH_DESC)*num);
+	  desc=xrealloc(desc,sizeof(KEYDB_SEARCH_DESC)*num);
 	}
     }
 
   if(count>0)
-    rc=keyserver_work(GET,NULL,desc,count,opt.keyserver);
+    rc=keyserver_work(KS_GET,NULL,desc,count,NULL,NULL,opt.keyserver);
 
-  m_free(desc);
+  xfree(desc);
 
   return rc;
 }
@@ -1516,7 +1649,9 @@ keyserver_import_fprint(const byte *fprint,size_t fprint_len,
 
   memcpy(desc.u.fpr,fprint,fprint_len);
 
-  return keyserver_work(GET,NULL,&desc,1,keyserver);
+  /* TODO: Warn here if the fingerprint we got doesn't match the one
+     we asked for? */
+  return keyserver_work(KS_GET,NULL,&desc,1,NULL,NULL,keyserver);
 }
 
 int 
@@ -1530,7 +1665,7 @@ keyserver_import_keyid(u32 *keyid,struct keyserver_spec *keyserver)
   desc.u.kid[0]=keyid[0];
   desc.u.kid[1]=keyid[1];
 
-  return keyserver_work(GET,NULL,&desc,1,keyserver);
+  return keyserver_work(KS_GET,NULL,&desc,1,NULL,NULL,keyserver);
 }
 
 /* code mostly stolen from do_export_stream */
@@ -1545,21 +1680,21 @@ keyidlist(STRLIST users,KEYDB_SEARCH_DESC **klist,int *count,int fakev3)
 
   *count=0;
 
-  *klist=m_alloc(sizeof(KEYDB_SEARCH_DESC)*num);
+  *klist=xmalloc(sizeof(KEYDB_SEARCH_DESC)*num);
 
   kdbhd=keydb_new(0);
 
   if(!users)
     {
       ndesc = 1;
-      desc = m_alloc_clear ( ndesc * sizeof *desc);
+      desc = xmalloc_clear ( ndesc * sizeof *desc);
       desc[0].mode = KEYDB_SEARCH_MODE_FIRST;
     }
   else
     {
       for (ndesc=0, sl=users; sl; sl = sl->next, ndesc++) 
 	;
-      desc = m_alloc ( ndesc * sizeof *desc);
+      desc = xmalloc ( ndesc * sizeof *desc);
         
       for (ndesc=0, sl=users; sl; sl = sl->next)
 	{
@@ -1604,7 +1739,7 @@ keyidlist(STRLIST users,KEYDB_SEARCH_DESC **klist,int *count,int fakev3)
 	      if(*count==num)
 		{
 		  num+=100;
-		  *klist=m_realloc(*klist,sizeof(KEYDB_SEARCH_DESC)*num);
+		  *klist=xrealloc(*klist,sizeof(KEYDB_SEARCH_DESC)*num);
 		}
 	    }
 
@@ -1660,7 +1795,7 @@ keyidlist(STRLIST users,KEYDB_SEARCH_DESC **klist,int *count,int fakev3)
 	      /* Try and parse the keyserver URL.  If it doesn't work,
 		 then we end up writing NULL which indicates we are
 		 the same as any other key. */
-	      if(uid && sig)
+	      if(sig)
 		(*klist)[*count].skipfncvalue=parse_preferred_keyserver(sig);
 	    }
 
@@ -1669,7 +1804,7 @@ keyidlist(STRLIST users,KEYDB_SEARCH_DESC **klist,int *count,int fakev3)
 	  if(*count==num)
 	    {
 	      num+=100;
-	      *klist=m_realloc(*klist,sizeof(KEYDB_SEARCH_DESC)*num);
+	      *klist=xrealloc(*klist,sizeof(KEYDB_SEARCH_DESC)*num);
 	    }
 	}
     }
@@ -1679,8 +1814,8 @@ keyidlist(STRLIST users,KEYDB_SEARCH_DESC **klist,int *count,int fakev3)
   
  leave:
   if(rc)
-    m_free(*klist);
-  m_free(desc);
+    xfree(*klist);
+  xfree(desc);
   keydb_release(kdbhd);
   release_kbnode(keyblock);
 
@@ -1690,7 +1825,7 @@ keyidlist(STRLIST users,KEYDB_SEARCH_DESC **klist,int *count,int fakev3)
 /* Note this is different than the original HKP refresh.  It allows
    usernames to refresh only part of the keyring. */
 
-int 
+int
 keyserver_refresh(STRLIST users)
 {
   int rc,count,numdesc,fakev3=0;
@@ -1724,7 +1859,7 @@ keyserver_refresh(STRLIST users)
       int i;
 
       /* Try to handle preferred keyserver keys first */
-      for(i=0;i<count;i++)
+      for(i=0;i<numdesc;i++)
 	{
 	  if(desc[i].skipfncvalue)
 	    {
@@ -1734,7 +1869,7 @@ keyserver_refresh(STRLIST users)
 		 Note that a preferred keyserver without a scheme://
 		 will be interpreted as hkp:// */
 
-	      rc=keyserver_work(GET,NULL,&desc[i],1,keyserver);
+	      rc=keyserver_work(KS_GET,NULL,&desc[i],1,NULL,NULL,keyserver);
 	      if(rc)
 		log_info(_("WARNING: unable to refresh key %s"
 			   " via %s: %s\n"),keystr_from_desc(&desc[i]),
@@ -1764,10 +1899,10 @@ keyserver_refresh(STRLIST users)
 		     count,opt.keyserver->uri);
 	}
 
-      rc=keyserver_work(GET,NULL,desc,numdesc,opt.keyserver);
+      rc=keyserver_work(KS_GET,NULL,desc,numdesc,NULL,NULL,opt.keyserver);
     }
 
-  m_free(desc);
+  xfree(desc);
 
   opt.keyserver_options.import_options=options;
 
@@ -1783,7 +1918,217 @@ int
 keyserver_search(STRLIST tokens)
 {
   if(tokens)
-    return keyserver_work(SEARCH,tokens,NULL,0,opt.keyserver);
+    return keyserver_work(KS_SEARCH,tokens,NULL,0,NULL,NULL,opt.keyserver);
   else
     return 0;
 }
+
+int
+keyserver_fetch(STRLIST urilist)
+{
+  KEYDB_SEARCH_DESC desc;
+  STRLIST sl;
+  unsigned int options=opt.keyserver_options.import_options;
+
+  /* Switch on fast-import, since fetch can handle more than one
+     import and we don't want each set to rebuild the trustdb.
+     Instead we do it once at the end. */
+  opt.keyserver_options.import_options|=IMPORT_FAST;
+
+  /* A dummy desc since we're not actually fetching a particular key
+     ID */
+  memset(&desc,0,sizeof(desc));
+  desc.mode=KEYDB_SEARCH_MODE_EXACT;
+
+  for(sl=urilist;sl;sl=sl->next)
+    {
+      struct keyserver_spec *spec;
+
+      spec=parse_keyserver_uri(sl->d,1,NULL,0);
+      if(spec)
+	{
+	  int rc;
+
+	  /*
+	    Set the direct_uri flag so we know later to call a direct
+	    handler instead of the keyserver style.  This lets us use
+	    gpgkeys_curl or gpgkeys_ldapuri instead of gpgkeys_ldap to
+	    fetch things like
+	    ldap://keyserver.pgp.com/o=PGP%20keys?pgpkey?sub?pgpkeyid=99242560
+	  */
+	  spec->flags.direct_uri=1;
+
+	  rc=keyserver_work(KS_GET,NULL,&desc,1,NULL,NULL,spec);
+	  if(rc)
+	    log_info (_("WARNING: unable to fetch URI %s: %s\n"),
+		     sl->d,g10_errstr(rc));
+
+	  free_keyserver_spec(spec);
+	}
+      else
+	log_info (_("WARNING: unable to parse URI %s\n"),sl->d);
+    }
+
+  opt.keyserver_options.import_options=options;
+
+  /* If the original options didn't have fast import, and the trustdb
+     is dirty, rebuild. */
+  if(!(opt.keyserver_options.import_options&IMPORT_FAST))
+    trustdb_check_or_update();
+
+  return 0;
+}
+
+/* Import key in a CERT or pointed to by a CERT */
+int
+keyserver_import_cert(const char *name,unsigned char **fpr,size_t *fpr_len)
+{
+  char *domain,*look,*url;
+  IOBUF key;
+  int type,rc=G10ERR_GENERAL;
+
+  look=xstrdup(name);
+
+  domain=strrchr(look,'@');
+  if(domain)
+    *domain='.';
+
+  type=get_cert(look,max_cert_size,&key,fpr,fpr_len,&url);
+  if(type==1)
+    {
+      int armor_status=opt.no_armor;
+
+      /* CERTs are always in binary format */
+      opt.no_armor=1;
+
+      rc=import_keys_stream(key,NULL,fpr,fpr_len,
+			    opt.keyserver_options.import_options);
+
+      opt.no_armor=armor_status;
+
+      iobuf_close(key);
+    }
+  else if(type==2 && *fpr)
+    {
+      /* We only consider the IPGP type if a fingerprint was provided.
+	 This lets us select the right key regardless of what a URL
+	 points to, or get the key from a keyserver. */
+      if(url)
+	{
+	  struct keyserver_spec *spec;
+
+	  spec=parse_keyserver_uri(url,1,NULL,0);
+	  if(spec)
+	    {
+	      STRLIST list=NULL;
+
+	      add_to_strlist(&list,url);
+
+	      rc=keyserver_fetch(list);
+
+	      free_strlist(list);
+	      free_keyserver_spec(spec);
+	    }
+	}
+      else if(opt.keyserver)
+	{
+	  /* If only a fingerprint is provided, try and fetch it from
+	     our --keyserver */
+
+	  rc=keyserver_import_fprint(*fpr,*fpr_len,opt.keyserver);
+	}
+
+      xfree(url);
+    }
+
+  xfree(look);
+
+  return rc;
+}
+
+/* Import key pointed to by a PKA record. Return the requested
+   fingerprint in fpr. */
+int
+keyserver_import_pka(const char *name,unsigned char **fpr,size_t *fpr_len)
+{
+  char *uri;
+  int rc=-1;
+
+  *fpr=xmalloc(20);
+  *fpr_len=20;
+
+  uri = get_pka_info (name, *fpr);
+  if (uri)
+    {
+      struct keyserver_spec *spec;
+      spec = parse_keyserver_uri (uri, 1, NULL, 0);
+      if (spec)
+	{
+	  rc=keyserver_import_fprint (*fpr, 20, spec);
+	  free_keyserver_spec (spec);
+	}
+      xfree (uri);
+    }
+
+  if(rc!=0)
+    xfree(*fpr);
+
+  return rc;
+}
+
+/* Import all keys that match name */
+int
+keyserver_import_name(const char *name,unsigned char **fpr,size_t *fpr_len,
+		      struct keyserver_spec *keyserver)
+{
+  STRLIST list=NULL;
+  int rc;
+
+  append_to_strlist(&list,name);
+
+  rc=keyserver_work(KS_GETNAME,list,NULL,0,fpr,fpr_len,keyserver);
+
+  free_strlist(list);
+
+  return rc;
+}
+
+/* Use the PGP Universal trick of asking ldap://keys.(maildomain) for
+   the key. */
+int
+keyserver_import_ldap(const char *name,unsigned char **fpr,size_t *fpr_len)
+{
+  char *domain;
+  struct keyserver_spec *keyserver;
+  STRLIST list=NULL;
+  int rc;
+
+  append_to_strlist(&list,name);
+
+  /* Parse out the domain */
+  domain=strrchr(name,'@');
+  if(!domain)
+    return G10ERR_GENERAL;
+
+  domain++;
+
+  keyserver=xmalloc_clear(sizeof(struct keyserver_spec));
+
+  keyserver->scheme=xstrdup("ldap");
+  keyserver->host=xmalloc(5+strlen(domain)+1);
+  strcpy(keyserver->host,"keys.");
+  strcat(keyserver->host,domain);
+  keyserver->uri=xmalloc(strlen(keyserver->scheme)+
+			 3+strlen(keyserver->host)+1);
+  strcpy(keyserver->uri,keyserver->scheme);
+  strcat(keyserver->uri,"://");
+  strcat(keyserver->uri,keyserver->host);
+    
+  rc=keyserver_work(KS_GETNAME,list,NULL,0,fpr,fpr_len,keyserver);
+
+  free_strlist(list);
+
+  free_keyserver_spec(keyserver);
+
+  return rc;
+}
