@@ -1,5 +1,5 @@
 /* gpgkeys_ldap.c - talk to a LDAP keyserver
- * Copyright (C) 2001, 2002, 2004, 2005 Free Software Foundation, Inc.
+ * Copyright (C) 2001, 2002, 2004, 2005, 2006 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -37,6 +37,8 @@
 #ifdef NEED_LBER_H
 #include <lber.h>
 #endif
+/* For OpenLDAP, to enable the API that we're using. */
+#define LDAP_DEPRECATED 1
 #include <ldap.h>
 #endif
 
@@ -248,7 +250,7 @@ join_two_modlists(LDAPMod ***one,LDAPMod **two)
    the attribute in question exists or not. */
 
 static int
-make_one_attr(LDAPMod ***modlist,int unique,char *attr,const char *value)
+make_one_attr(LDAPMod ***modlist,char *attr,const char *value)
 {
   LDAPMod **m;
   int nummods=0;
@@ -270,7 +272,8 @@ make_one_attr(LDAPMod ***modlist,int unique,char *attr,const char *value)
 	  if(ptr)
 	    for(ptr=(*m)->mod_values;*ptr;ptr++)
 	      {
-		if(unique && strcmp(*ptr,value)==0)
+		/* Duplicate value */
+		if(strcmp(*ptr,value)==0)
 		  return 1;
 		numvalues++;
 	      }
@@ -363,8 +366,8 @@ build_attrs(LDAPMod ***modlist,char *line)
 
       if(strlen(tok)==16)
 	{
-	  make_one_attr(modlist,0,"pgpCertID",tok);
-	  make_one_attr(modlist,0,"pgpKeyID",&tok[8]);
+	  make_one_attr(modlist,"pgpCertID",tok);
+	  make_one_attr(modlist,"pgpKeyID",&tok[8]);
 	}
       else
 	return;
@@ -376,11 +379,11 @@ build_attrs(LDAPMod ***modlist,char *line)
       switch(atoi(tok))
 	{
 	case 1:
-	  make_one_attr(modlist,0,"pgpKeyType","RSA");
+	  make_one_attr(modlist,"pgpKeyType","RSA");
 	  break;
 
 	case 17:
-	  make_one_attr(modlist,0,"pgpKeyType","DSS/DH");
+	  make_one_attr(modlist,"pgpKeyType","DSS/DH");
 	  break;
 	}
 
@@ -398,7 +401,7 @@ build_attrs(LDAPMod ***modlist,char *line)
 	  if(val<99999 && val>0)
 	    {
 	      sprintf(padded,"%05u",atoi(tok));
-	      make_one_attr(modlist,0,"pgpKeySize",padded);
+	      make_one_attr(modlist,"pgpKeySize",padded);
 	    }
 	}
 
@@ -411,7 +414,7 @@ build_attrs(LDAPMod ***modlist,char *line)
 	  char *stamp=epoch2ldaptime(atoi(tok));
 	  if(stamp)
 	    {
-	      make_one_attr(modlist,0,"pgpKeyCreateTime",stamp);
+	      make_one_attr(modlist,"pgpKeyCreateTime",stamp);
 	      free(stamp);
 	    }
 	}
@@ -425,7 +428,7 @@ build_attrs(LDAPMod ***modlist,char *line)
 	  char *stamp=epoch2ldaptime(atoi(tok));
 	  if(stamp)
 	    {
-	      make_one_attr(modlist,0,"pgpKeyExpireTime",stamp);
+	      make_one_attr(modlist,"pgpKeyExpireTime",stamp);
 	      free(stamp);
 	    }
 	}
@@ -455,8 +458,8 @@ build_attrs(LDAPMod ***modlist,char *line)
 	"(&(pgpUserID=*isabella*)(pgpDisabled=0))"
       */
 
-      make_one_attr(modlist,0,"pgpDisabled",disabled?"1":"0");
-      make_one_attr(modlist,0,"pgpRevoked",revoked?"1":"0");
+      make_one_attr(modlist,"pgpDisabled",disabled?"1":"0");
+      make_one_attr(modlist,"pgpRevoked",revoked?"1":"0");
     }
   else if(ascii_strcasecmp("sub",record)==0)
     {
@@ -467,7 +470,7 @@ build_attrs(LDAPMod ***modlist,char *line)
 	return;
 
       if(strlen(tok)==16)
-	make_one_attr(modlist,0,"pgpSubKeyID",tok);
+	make_one_attr(modlist,"pgpSubKeyID",tok);
       else
 	return;
 
@@ -489,7 +492,7 @@ build_attrs(LDAPMod ***modlist,char *line)
 	  if(val<99999 && val>0)
 	    {
 	      sprintf(padded,"%05u",atoi(tok));
-	      make_one_attr(modlist,0,"pgpKeySize",padded);
+	      make_one_attr(modlist,"pgpKeySize",padded);
 	    }
 	}
 
@@ -531,7 +534,7 @@ build_attrs(LDAPMod ***modlist,char *line)
       /* We don't care about the other info provided in the uid: line
 	 since the LDAP schema doesn't need it. */
 
-      make_one_attr(modlist,0,"pgpUserID",userid);
+      make_one_attr(modlist,"pgpUserID",userid);
     }
   else if(ascii_strcasecmp("sig",record)==0)
     {
@@ -541,7 +544,7 @@ build_attrs(LDAPMod ***modlist,char *line)
 	return;
 
       if(strlen(tok)==16)
-	make_one_attr(modlist,1,"pgpSignerID",tok);
+	make_one_attr(modlist,"pgpSignerID",tok);
     }
 }
 
@@ -590,17 +593,17 @@ send_key(int *eof)
   /* Start by nulling out all attributes.  We try and do a modify
      operation first, so this ensures that we don't leave old
      attributes lying around. */
-  make_one_attr(&modlist,0,"pgpDisabled",NULL);
-  make_one_attr(&modlist,0,"pgpKeyID",NULL);
-  make_one_attr(&modlist,0,"pgpKeyType",NULL);
-  make_one_attr(&modlist,0,"pgpUserID",NULL);
-  make_one_attr(&modlist,0,"pgpKeyCreateTime",NULL);
-  make_one_attr(&modlist,0,"pgpSignerID",NULL);
-  make_one_attr(&modlist,0,"pgpRevoked",NULL);
-  make_one_attr(&modlist,0,"pgpSubKeyID",NULL);
-  make_one_attr(&modlist,0,"pgpKeySize",NULL);
-  make_one_attr(&modlist,0,"pgpKeyExpireTime",NULL);
-  make_one_attr(&modlist,0,"pgpCertID",NULL);
+  make_one_attr(&modlist,"pgpDisabled",NULL);
+  make_one_attr(&modlist,"pgpKeyID",NULL);
+  make_one_attr(&modlist,"pgpKeyType",NULL);
+  make_one_attr(&modlist,"pgpUserID",NULL);
+  make_one_attr(&modlist,"pgpKeyCreateTime",NULL);
+  make_one_attr(&modlist,"pgpSignerID",NULL);
+  make_one_attr(&modlist,"pgpRevoked",NULL);
+  make_one_attr(&modlist,"pgpSubKeyID",NULL);
+  make_one_attr(&modlist,"pgpKeySize",NULL);
+  make_one_attr(&modlist,"pgpKeyExpireTime",NULL);
+  make_one_attr(&modlist,"pgpCertID",NULL);
 
   /* Assemble the INFO stuff into LDAP attributes */
 
@@ -719,8 +722,8 @@ send_key(int *eof)
       goto fail;
     }
 
-  make_one_attr(&addlist,0,"objectClass","pgpKeyInfo");
-  make_one_attr(&addlist,0,"pgpKey",key);
+  make_one_attr(&addlist,"objectClass","pgpKeyInfo");
+  make_one_attr(&addlist,"pgpKey",key);
 
   /* Now append addlist onto modlist */
   if(!join_two_modlists(&modlist,addlist))
@@ -1116,6 +1119,152 @@ get_key(char *getkey)
   return ret;
 }
 
+#define LDAP_ESCAPE_CHARS "*()\\"
+
+static int
+ldap_quote(char *buffer,const char *string)
+{
+  int count=0;
+
+  for(;*string;string++)
+    {
+      if(strchr(LDAP_ESCAPE_CHARS,*string))
+	{
+	  if(buffer)
+	    {
+	      sprintf(buffer,"\\%02X",*string);
+	      buffer+=3;
+	    }
+
+	  count+=3;
+	}
+      else
+	{
+	  if(buffer)
+	    *buffer++=*string;
+
+	  count++;
+	}
+    }
+
+  if(buffer)
+    *buffer='\0';
+
+  return count;
+}
+
+/* Note that key-not-found is not a fatal error */
+static int
+get_name(char *getkey)
+{
+  LDAPMessage *res,*each;
+  int ret=KEYSERVER_INTERNAL_ERROR,err,count;
+  char *expanded_search;
+  /* The maximum size of the search, including the optional stuff and
+     the trailing \0 */
+  char search[2+11+3+MAX_LINE+2+15+14+1+1+20];
+  /* This ordering is significant - specifically, "pgpcertid" needs to
+     be the second item in the list, since everything after it may be
+     discarded if the user isn't in verbose mode. */
+  char *attrs[]={"replaceme","pgpcertid","pgpuserid","pgpkeyid","pgprevoked",
+		 "pgpdisabled","pgpkeycreatetime","modifytimestamp",
+		 "pgpkeysize","pgpkeytype",NULL};
+  attrs[0]=pgpkeystr; /* Some compilers don't like using variables as
+                         array initializers. */
+
+  expanded_search=malloc(ldap_quote(NULL,getkey)+1);
+  if(!expanded_search)
+    {
+      fprintf(output,"NAME %s FAILED %d\n",getkey,KEYSERVER_NO_MEMORY);
+      fprintf(console,"Out of memory when quoting LDAP search string\n");
+      return KEYSERVER_NO_MEMORY;
+    }
+
+  ldap_quote(expanded_search,getkey);
+
+  /* Build the search string */
+
+  sprintf(search,"%s(pgpuserid=*%s*)%s%s%s",
+	  (!(opt->flags.include_disabled&&opt->flags.include_revoked))?"(&":"",
+	  expanded_search,
+	  opt->flags.include_disabled?"":"(pgpdisabled=0)",
+	  opt->flags.include_revoked?"":"(pgprevoked=0)",
+	  !(opt->flags.include_disabled&&opt->flags.include_revoked)?")":"");
+
+  free(expanded_search);
+
+  if(opt->verbose>2)
+    fprintf(console,"gpgkeys: LDAP fetch for: %s\n",search);
+
+  if(!opt->verbose)
+    attrs[2]=NULL; /* keep only pgpkey(v2) and pgpcertid */
+
+  err=ldap_search_s(ldap,basekeyspacedn,
+		    LDAP_SCOPE_SUBTREE,search,attrs,0,&res);
+  if(err!=0)
+    {
+      int errtag=ldap_err_to_gpg_err(err);
+
+      fprintf(console,"gpgkeys: LDAP search error: %s\n",ldap_err2string(err));
+      fprintf(output,"NAME %s BEGIN\n",getkey);
+      fprintf(output,"NAME %s FAILED %d\n",getkey,errtag);
+      return errtag;
+    }
+
+  count=ldap_count_entries(ldap,res);
+  if(count<1)
+    {
+      fprintf(console,"gpgkeys: key %s not found on keyserver\n",getkey);
+      fprintf(output,"NAME %s BEGIN\n",getkey);
+      fprintf(output,"NAME %s FAILED %d\n",getkey,KEYSERVER_KEY_NOT_FOUND);
+    }
+  else
+    {
+      /* There may be more than one result, but we return them all. */
+
+      each=ldap_first_entry(ldap,res);
+      while(each!=NULL)
+	{
+	  char **vals,**certid;
+
+	  certid=ldap_get_values(ldap,each,"pgpcertid");
+	  if(certid!=NULL)
+	    {
+	      build_info(certid[0],each);
+
+	      fprintf(output,"NAME %s BEGIN\n",getkey);
+
+	      vals=ldap_get_values(ldap,each,pgpkeystr);
+	      if(vals==NULL)
+		{
+		  int errtag=ldap_to_gpg_err(ldap);
+
+		  fprintf(console,"gpgkeys: unable to retrieve key %s "
+			  "from keyserver\n",getkey);
+		  fprintf(output,"NAME %s FAILED %d\n",getkey,errtag);
+		}
+	      else
+		{
+		  print_nocr(output,vals[0]);
+		  fprintf(output,"\nNAME %s END\n",getkey);
+
+		  ldap_value_free(vals);
+		}
+
+	      ldap_value_free(certid);
+	    }
+
+	  each=ldap_next_entry(ldap,each);
+	}
+    }
+
+  ret=KEYSERVER_OK;
+
+  ldap_msgfree(res);
+
+  return ret;
+}
+
 static void
 printquoted(FILE *stream,char *string,char delim)
 {
@@ -1133,30 +1282,55 @@ printquoted(FILE *stream,char *string,char delim)
 /* Returns 0 on success and -1 on error.  Note that key-not-found is
    not an error! */
 static int
-search_key(char *searchkey)
+search_key(const char *searchkey)
 {
   char **vals;
   LDAPMessage *res,*each;
   int err,count=0;
   struct keylist *dupelist=NULL;
+  char *expanded_search;
   /* The maximum size of the search, including the optional stuff and
      the trailing \0 */
-  char search[2+12+MAX_LINE+2+15+14+1+1];
+  char search[2+11+3+MAX_LINE+2+15+14+1+1+20];
   char *attrs[]={"pgpcertid","pgpuserid","pgprevoked","pgpdisabled",
 		 "pgpkeycreatetime","pgpkeyexpiretime","modifytimestamp",
 		 "pgpkeysize","pgpkeytype",NULL};
+  enum ks_search_type search_type;
 
   fprintf(output,"SEARCH %s BEGIN\n",searchkey);
 
+  search_type=classify_ks_search(&searchkey);
+
+  if(opt->debug)
+    fprintf(console,"search type is %d, and key is \"%s\"\n",
+	    search_type,searchkey);
+
+  expanded_search=malloc(ldap_quote(NULL,searchkey)+1);
+  if(!expanded_search)
+    {
+      fprintf(output,"SEARCH %s FAILED %d\n",searchkey,KEYSERVER_NO_MEMORY);
+      fprintf(console,"Out of memory when quoting LDAP search string\n");
+      return KEYSERVER_NO_MEMORY;
+    }
+
+  ldap_quote(expanded_search,searchkey);
+
   /* Build the search string */
 
-  sprintf(search,"%s(pgpuserid=*%s*)%s%s%s",
+  sprintf(search,"%s(pgpuserid=%s%s%s)%s%s%s",
 	  (!(opt->flags.include_disabled&&opt->flags.include_revoked))?"(&":"",
-	  searchkey,
+	  (search_type==KS_SEARCH_EXACT)?"":
+	  (search_type==KS_SEARCH_MAILSUB)?"*<*":"*",
+	  expanded_search,
+	  (search_type==KS_SEARCH_EXACT
+	   || search_type==KS_SEARCH_MAIL)?"":
+	  (search_type==KS_SEARCH_MAILSUB)?"*>":"*",
 	  opt->flags.include_disabled?"":"(pgpdisabled=0)",
 	  opt->flags.include_revoked?"":"(pgprevoked=0)",
 	  !(opt->flags.include_disabled&&opt->flags.include_revoked)?")":"");
 
+  free(expanded_search);
+
   if(opt->verbose>2)
     fprintf(console,"gpgkeys: LDAP search for: %s\n",search);
 
@@ -1198,7 +1372,14 @@ search_key(char *searchkey)
     }
 
   if(err==LDAP_SIZELIMIT_EXCEEDED)
-    fprintf(console,"gpgkeys: search results exceeded server limit.  First %d results shown.\n",count);
+    {
+      if(count==1)
+	fprintf(console,"gpgkeys: search results exceeded server limit."
+		"  First %d result shown.\n",count);
+      else
+	fprintf(console,"gpgkeys: search results exceeded server limit."
+		"  First %d results shown.\n",count);
+    }
 
   free_keylist(dupelist);
   dupelist=NULL;
@@ -1536,7 +1717,7 @@ int
 main(int argc,char *argv[])
 {
   int port=0,arg,err,ret=KEYSERVER_INTERNAL_ERROR;
-  char line[MAX_LINE];
+  char line[MAX_LINE],*binddn=NULL,*bindpw=NULL;
   int failed=0,use_ssl=0,use_tls=0,bound=0;
   struct keylist *keylist=NULL,*keyptr=NULL;
 
@@ -1672,6 +1853,50 @@ main(int argc,char *argv[])
 		  real_ldap=1;
 		}
 	    }
+	  else if(strncasecmp(start,"binddn",6)==0)
+	    {
+	      if(no)
+		{
+		  free(binddn);
+		  binddn=NULL;
+		}
+	      else if(start[6]=='=')
+		{
+		  free(binddn);
+		  binddn=strdup(&start[7]);
+		  if(!binddn)
+		    {
+		      fprintf(console,"gpgkeys: out of memory while creating "
+			      "bind DN\n");
+		      ret=KEYSERVER_NO_MEMORY;
+		      goto fail;
+		    }
+
+		  real_ldap=1;
+		}
+	    }
+	  else if(strncasecmp(start,"bindpw",6)==0)
+	    {
+	      if(no)
+		{
+		  free(bindpw);
+		  bindpw=NULL;
+		}
+	      else if(start[6]=='=')
+		{
+		  free(bindpw);
+		  bindpw=strdup(&start[7]);
+		  if(!bindpw)
+		    {
+		      fprintf(console,"gpgkeys: out of memory while creating "
+			      "bind password\n");
+		      ret=KEYSERVER_NO_MEMORY;
+		      goto fail;
+		    }
+
+		  real_ldap=1;
+		}
+	    }
 
 	  continue;
 	}
@@ -1705,7 +1930,8 @@ main(int argc,char *argv[])
       return KEYSERVER_INTERNAL_ERROR;
     }
 
-#if defined(HAVE_LDAP_SET_OPTION) && defined(LDAP_OPT_X_TLS_CACERTFILE)
+#if defined(LDAP_OPT_X_TLS_CACERTFILE) && defined(HAVE_LDAP_SET_OPTION)
+
   if(opt->ca_cert_file)
     {
       err=ldap_set_option(NULL,LDAP_OPT_X_TLS_CACERTFILE,opt->ca_cert_file);
@@ -1717,7 +1943,7 @@ main(int argc,char *argv[])
 	  goto fail;
 	}
     }
-#endif /* HAVE_LDAP_SET_OPTION && LDAP_OPT_X_TLS_CACERTFILE */
+#endif /* LDAP_OPT_X_TLS_CACERTFILE && HAVE_LDAP_SET_OPTION */
 
   /* SSL trumps TLS */
   if(use_ssl)
@@ -1728,7 +1954,8 @@ main(int argc,char *argv[])
 
   if(opt->action==KS_SEND)
     while(fgets(line,MAX_LINE,input)!=NULL && line[0]!='\n');
-  else if(opt->action==KS_GET || opt->action==KS_SEARCH)
+  else if(opt->action==KS_GET
+	  || opt->action==KS_GETNAME || opt->action==KS_SEARCH)
     {
       for(;;)
 	{
@@ -1880,6 +2107,8 @@ main(int argc,char *argv[])
 	  int ver=LDAP_VERSION3;
 
 	  err=ldap_set_option(ldap,LDAP_OPT_PROTOCOL_VERSION,&ver);
+
+#ifdef LDAP_OPT_X_TLS
 	  if(err==LDAP_SUCCESS)
 	    {
 	      if(opt->flags.check_cert)
@@ -1888,9 +2117,11 @@ main(int argc,char *argv[])
 		ver=LDAP_OPT_X_TLS_NEVER;
 
 	      err=ldap_set_option(ldap,LDAP_OPT_X_TLS_REQUIRE_CERT,&ver);
-	      if(err==LDAP_SUCCESS)
-		err=ldap_start_tls_s(ldap,NULL,NULL);
 	    }
+#endif
+
+	  if(err==LDAP_SUCCESS)
+	    err=ldap_start_tls_s(ldap,NULL,NULL);
 
 	  if(err!=LDAP_SUCCESS)
 	    {
@@ -1904,7 +2135,7 @@ main(int argc,char *argv[])
 		  goto fail;
 		}
 	    }
-	  else if(err==LDAP_SUCCESS && opt->verbose>1)
+	  else if(opt->verbose>1)
 	    fprintf(console,"gpgkeys: TLS started successfully.\n");
 #else
 	  if(use_tls>=2)
@@ -1919,25 +2150,40 @@ main(int argc,char *argv[])
 	}
     }
 
-#if 0
-  /* The LDAP keyserver doesn't require this, but it might be useful
-     if someone stores keys on a V2 LDAP server somewhere.  (V3
-     doesn't require a bind).  Leave this out for now since it is not
-     clear if anyone's server we're likely to use really cares, plus
-     there are some servers that don't allow it. */
+  /* By default we don't bind as there is usually no need to.  For
+     cases where the server needs some authentication, the user can
+     use binddn and bindpw for auth. */
 
-  err=ldap_simple_bind_s(ldap,NULL,NULL);
-  if(err!=0)
+  if(binddn)
     {
-      fprintf(console,"gpgkeys: internal LDAP bind error: %s\n",
-	      ldap_err2string(err));
-      fail_all(keylist,ldap_err_to_gpg_err(err));
-      goto fail;
-    }
-  else
-    bound=1;
+#ifdef HAVE_LDAP_SET_OPTION
+      int ver=LDAP_VERSION3;
+
+      err=ldap_set_option(ldap,LDAP_OPT_PROTOCOL_VERSION,&ver);
+      if(err!=LDAP_SUCCESS)
+	{
+	  fprintf(console,"gpgkeys: unable to go to LDAP 3: %s\n",
+		  ldap_err2string(err));
+	  fail_all(keylist,ldap_err_to_gpg_err(err));
+	  goto fail;
+	}
 #endif
 
+      if(opt->verbose>2)
+	fprintf(console,"gpgkeys: LDAP bind to %s, pw %s\n",binddn,
+		bindpw?">not shown<":">none<");
+      err=ldap_simple_bind_s(ldap,binddn,bindpw);
+      if(err!=LDAP_SUCCESS)
+	{
+	  fprintf(console,"gpgkeys: internal LDAP bind error: %s\n",
+		  ldap_err2string(err));
+	  fail_all(keylist,ldap_err_to_gpg_err(err));
+	  goto fail;
+	}
+      else
+	bound=1;
+    }
+
   if(opt->action==KS_GET)
     {
       keyptr=keylist;
@@ -1949,6 +2195,20 @@ main(int argc,char *argv[])
 	  if(get_key(keyptr->str)!=KEYSERVER_OK)
 	    failed++;
 
+	  keyptr=keyptr->next;
+	}
+    }
+  else if(opt->action==KS_GETNAME)
+    {
+      keyptr=keylist;
+
+      while(keyptr!=NULL)
+	{
+	  set_timeout(opt->timeout);
+
+	  if(get_name(keyptr->str)!=KEYSERVER_OK)
+	    failed++;
+
 	  keyptr=keyptr->next;
 	}
     }
