@@ -36,7 +36,7 @@
    strlen("OPAQUE")+1+sizeof_opaque+1 */
 #define MAX_LINE       (6+1+1024+1)
 
-#define MAX_COMMAND    6
+#define MAX_COMMAND    7
 #define MAX_OPTION   256
 #define MAX_SCHEME    20
 #define MAX_OPAQUE  1024
@@ -72,7 +72,10 @@ struct keylist
 unsigned int set_timeout(unsigned int seconds);
 int register_timeout(void);
 
-enum ks_action {KS_UNKNOWN=0,KS_GET,KS_SEND,KS_SEARCH};
+enum ks_action {KS_UNKNOWN=0,KS_GET,KS_GETNAME,KS_SEND,KS_SEARCH};
+
+enum ks_search_type {KS_SEARCH_SUBSTR,KS_SEARCH_EXACT,
+		     KS_SEARCH_MAIL,KS_SEARCH_MAILSUB};
 
 struct ks_options
 {
@@ -101,18 +104,31 @@ void free_ks_options(struct ks_options *opt);
 int parse_ks_options(char *line,struct ks_options *opt);
 const char *ks_action_to_string(enum ks_action action);
 void print_nocr(FILE *stream,const char *str);
+enum ks_search_type classify_ks_search(const char **search);
 
 #if defined (HAVE_LIBCURL) || defined (FAKE_CURL)
 int curl_err_to_gpg_err(CURLcode error);
 
 struct curl_writer_ctx
 {
-  int initialized,markeridx,begun,done;
+  struct
+  {
+    unsigned int initialized:1;
+    unsigned int begun:1;
+    unsigned int done:1;
+    unsigned int armor:1;
+  } flags;
+
+  int armor_remaining;
+  unsigned char armor_ctx[3];
+  int markeridx,linelen;
   const char *marker;
   FILE *stream;
 };
 
 size_t curl_writer(const void *ptr,size_t size,size_t nmemb,void *cw_ctx);
+void curl_writer_finalize(struct curl_writer_ctx *ctx);
+
 #endif
 
 #endif /* !_KSUTIL_H_ */
