@@ -1,5 +1,5 @@
 /* memory.c  -	memory allocation
- *	Copyright (C) 1998, 1999, 2001 Free Software Foundation, Inc.
+ *	Copyright (C) 1998, 1999, 2001, 2005 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -15,7 +15,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  *
  *
  * We use our own memory allocation functions instead of plain malloc(),
@@ -48,9 +49,9 @@
  * replace this guard stuff with one provided by a modern malloc library
  */
 #if SIZEOF_UNSIGNED_LONG == 8
-  #define EXTRA_ALIGN 4
+#define EXTRA_ALIGN 4
 #else
-  #define EXTRA_ALIGN 0
+#define EXTRA_ALIGN 0
 #endif
 
 #if defined(M_DEBUG) || defined(M_GUARD)
@@ -59,32 +60,36 @@
 
 #ifdef M_DEBUG
 
-  #ifndef M_GUARD
-    #define M_GUARD 1
-  #endif
-  #undef m_alloc
-  #undef m_alloc_clear
-  #undef m_alloc_secure
-  #undef m_alloc_secure_clear
-  #undef m_realloc
-  #undef m_free
-  #undef m_check
-  #undef m_strdup
-  #define FNAME(a)  m_debug_ ##a
-  #define FNAMEPRT  , const char *info
-  #define FNAMEARG  , info
-  #ifndef __riscos__
-    #define store_len(p,n,m) do { add_entry(p,n,m, \
+#ifndef M_GUARD
+#define M_GUARD 1
+#endif
+#undef xmalloc
+#undef xmalloc_clear
+#undef xmalloc_secure
+#undef xmalloc_secure_clear
+#undef xrealloc
+#undef xfree
+#undef m_check
+#undef xstrdup
+#define FNAME(a)   m_debug_ ##a
+#define FNAMEX(a)  m_debug_ ##a
+#define FNAMEXM(a) m_debug_ ##a
+#define FNAMEPRT  , const char *info
+#define FNAMEARG  , info
+#ifndef __riscos__
+#define store_len(p,n,m) do { add_entry(p,n,m, \
 					info, __FUNCTION__);  } while(0)
-  #else
-    #define store_len(p,n,m) do { add_entry(p,n,m, \
-	          info, "[" __FILE__ ":" STR(__LINE__) "]" );  } while(0)
-  #endif
 #else
-  #define FNAME(a)  m_ ##a
-  #define FNAMEPRT
-  #define FNAMEARG
-  #define store_len(p,n,m) do { ((byte*)p)[EXTRA_ALIGN+0] = n;		      \
+#define store_len(p,n,m) do { add_entry(p,n,m, \
+	          info, __func__ );  } while(0)
+#endif
+#else
+#define FNAME(a)   m_ ##a
+#define FNAMEX(a)  x ##a
+#define FNAMEXM(a) xm ##a
+#define FNAMEPRT
+#define FNAMEARG
+#define store_len(p,n,m) do { ((byte*)p)[EXTRA_ALIGN+0] = n;		      \
 				((byte*)p)[EXTRA_ALIGN+1] = n >> 8 ;	      \
 				((byte*)p)[EXTRA_ALIGN+2] = n >> 16 ;	      \
 				((byte*)p)[EXTRA_ALIGN+3] = m? MAGIC_SEC_BYTE \
@@ -175,7 +180,7 @@ add_entry( byte *p, unsigned n, int mode, const char *info, const char *by )
     e = memtbl+index;
     if( e->inuse )
 	membug("Ooops: entry %u is flagged as in use\n", index);
-    e->user_p = p + 4;
+    e->user_p = p + EXTRA_ALIGN + 4;
     e->user_n = n;
     e->count++;
     if( e->next )
@@ -197,10 +202,10 @@ add_entry( byte *p, unsigned n, int mode, const char *info, const char *by )
     e->inuse = 1;
 
     /* put the index at the start of the memory */
-    p[0] = index;
-    p[1] = index >> 8 ;
-    p[2] = index >> 16 ;
-    p[3] = mode? MAGIC_SEC_BYTE : MAGIC_NOR_BYTE  ;
+    p[EXTRA_ALIGN+0] = index;
+    p[EXTRA_ALIGN+1] = index >> 8 ;
+    p[EXTRA_ALIGN+2] = index >> 16 ;
+    p[EXTRA_ALIGN+3] = mode? MAGIC_SEC_BYTE : MAGIC_NOR_BYTE  ;
     if( DBG_MEMORY )
 	log_debug( "%s allocates %u bytes using %s\n", info, e->user_n, by );
 }
@@ -277,8 +282,12 @@ free_entry( byte *p, const char *info )
 	    ;
 	e2->next = e;
     }
-    memset(p,'f', e->user_n+5);
-    free(p);
+    if( m_is_secure(p+EXTRA_ALIGN+4) )
+	secmem_free(p);
+    else {
+        memset(p,'f', e->user_n+5);
+	free(p);
+    }
 }
 
 static void
@@ -328,11 +337,11 @@ check_allmem( const char *info )
 
     for( e = memtbl, n = 0; n < memtbl_len; n++, e++ ) {
 	if( e->inuse ) {
-          #ifndef __riscos__
+#ifndef __riscos__
 	    check_mem(e->user_p-4-EXTRA_ALIGN, info);
-          #else 
+#else 
 	    check_mem((const byte *) e->user_p-4-EXTRA_ALIGN, info);
-          #endif
+#endif
         }
     }
 }
@@ -350,10 +359,10 @@ membug( const char *fmt, ... )
     vfprintf(stderr,fmt,arg_ptr) ;
     va_end(arg_ptr);
     fflush(stderr);
-  #ifdef M_DEBUG
+#ifdef M_DEBUG
     if( DBG_MEMSTAT )
 	dump_table();
-  #endif
+#endif
     abort();
 }
 #endif
@@ -361,7 +370,7 @@ membug( const char *fmt, ... )
 void
 m_print_stats( const char *prefix )
 {
-  #ifdef M_DEBUG
+#ifdef M_DEBUG
     unsigned n;
     struct memtbl_entry *e;
     ulong sum = 0, chunks =0;
@@ -375,19 +384,19 @@ m_print_stats( const char *prefix )
 
     log_debug( "%s%smemstat: %8lu bytes in %ld chunks used\n",
 		prefix? prefix:"", prefix? ": ":"", sum, chunks );
-  #elif defined(M_GUARD)
+#elif defined(M_GUARD)
     log_debug( "%s%smemstat: %8ld bytes\n",
 		prefix? prefix:"", prefix? ": ":"", used_memory );
-  #endif
+#endif
 }
 
 void
 m_dump_table( const char *prefix )
 {
-  #ifdef M_DEBUG
+#ifdef M_DEBUG
     fprintf(stderr,"Memory-Table-Dump: %s\n", prefix);
     dump_table();
-  #endif
+#endif
     m_print_stats( prefix );
 }
 
@@ -402,6 +411,9 @@ out_of_core(size_t n, int secure)
         log_info ("(this may be caused by too many secret keys used "
                   "simultaneously or due to excessive large key sizes)\n");
     }
+#if defined(M_GUARD) && defined(__riscos__)
+    abort();
+#endif
     exit (2);
 }
 
@@ -410,22 +422,28 @@ out_of_core(size_t n, int secure)
  * This function gives up if we do not have enough memory
  */
 void *
-FNAME(alloc)( size_t n FNAMEPRT )
+FNAMEXM(alloc)( size_t n FNAMEPRT )
 {
     char *p;
 
-  #ifdef M_GUARD
+#ifdef M_GUARD
+    if(!n)
+      out_of_core(n,0); /* should never happen */
     if( !(p = malloc( n + EXTRA_ALIGN+5 )) )
 	out_of_core(n,0);
     store_len(p,n,0);
     used_memory += n;
     p[4+EXTRA_ALIGN+n] = MAGIC_END_BYTE;
     return p+EXTRA_ALIGN+4;
-  #else
+#else
+    /* mallocing zero bytes is undefined by ISO-C, so we better make
+       sure that it won't happen */
+    if (!n)
+      n = 1;
     if( !(p = malloc( n )) )
 	out_of_core(n,0);
     return p;
-  #endif
+#endif
 }
 
 /****************
@@ -433,37 +451,43 @@ FNAME(alloc)( size_t n FNAMEPRT )
  * This function gives up if we do not have enough memory
  */
 void *
-FNAME(alloc_secure)( size_t n FNAMEPRT )
+FNAMEXM(alloc_secure)( size_t n FNAMEPRT )
 {
     char *p;
 
-  #ifdef M_GUARD
+#ifdef M_GUARD
+    if(!n)
+      out_of_core(n,1); /* should never happen */
     if( !(p = secmem_malloc( n +EXTRA_ALIGN+ 5 )) )
 	out_of_core(n,1);
     store_len(p,n,1);
     p[4+EXTRA_ALIGN+n] = MAGIC_END_BYTE;
     return p+EXTRA_ALIGN+4;
-  #else
+#else
+    /* mallocing zero bytes is undefined by ISO-C, so we better make
+       sure that it won't happen */
+    if (!n)
+      n = 1;
     if( !(p = secmem_malloc( n )) )
 	out_of_core(n,1);
     return p;
-  #endif
+#endif
 }
 
 void *
-FNAME(alloc_clear)( size_t n FNAMEPRT )
+FNAMEXM(alloc_clear)( size_t n FNAMEPRT )
 {
     void *p;
-    p = FNAME(alloc)( n FNAMEARG );
+    p = FNAMEXM(alloc)( n FNAMEARG );
     memset(p, 0, n );
     return p;
 }
 
 void *
-FNAME(alloc_secure_clear)( size_t n FNAMEPRT)
+FNAMEXM(alloc_secure_clear)( size_t n FNAMEPRT)
 {
     void *p;
-    p = FNAME(alloc_secure)( n FNAMEARG );
+    p = FNAMEXM(alloc_secure)( n FNAMEARG );
     memset(p, 0, n );
     return p;
 }
@@ -473,11 +497,11 @@ FNAME(alloc_secure_clear)( size_t n FNAMEPRT)
  * realloc and clear the old space
  */
 void *
-FNAME(realloc)( void *a, size_t n FNAMEPRT )
+FNAMEX(realloc)( void *a, size_t n FNAMEPRT )
 {
     void *b;
 
-  #ifdef M_GUARD
+#ifdef M_GUARD
     if( a ) {
         unsigned char *p = a;
         size_t len = m_size(a);
@@ -493,17 +517,17 @@ FNAME(realloc)( void *a, size_t n FNAMEPRT )
         FNAME(free)(p FNAMEARG);
     }
     else
-        b = FNAME(alloc)(n);
-  #else
+        b = FNAME(alloc)(n FNAMEARG);
+#else
     if( m_is_secure(a) ) {
-	if( !(b = secmem_realloc( a, n )) )
+	if( !(b = secmexrealloc( a, n )) )
 	    out_of_core(n,1);
     }
     else {
 	if( !(b = realloc( a, n )) )
 	    out_of_core(n,0);
     }
-  #endif
+#endif
 
     return b;
 }
@@ -514,15 +538,15 @@ FNAME(realloc)( void *a, size_t n FNAMEPRT )
  * Free a pointer
  */
 void
-FNAME(free)( void *a FNAMEPRT )
+FNAMEX(free)( void *a FNAMEPRT )
 {
     byte *p = a;
 
     if( !p )
 	return;
-  #ifdef M_DEBUG
+#ifdef M_DEBUG
     free_entry(p-EXTRA_ALIGN-4, info);
-  #elif defined M_GUARD
+#elif defined M_GUARD
     m_check(p);
     if( m_is_secure(a) )
 	secmem_free(p-EXTRA_ALIGN-4);
@@ -530,89 +554,93 @@ FNAME(free)( void *a FNAMEPRT )
 	used_memory -= m_size(a);
 	free(p-EXTRA_ALIGN-4);
     }
-  #else
+#else
     if( m_is_secure(a) )
 	secmem_free(p);
     else
 	free(p);
-  #endif
+#endif
 }
 
 
 void
 FNAME(check)( const void *a FNAMEPRT )
 {
-  #ifdef M_GUARD
+#ifdef M_GUARD
     const byte *p = a;
 
-  #ifdef M_DEBUG
+#ifdef M_DEBUG
     if( p )
 	check_mem(p-EXTRA_ALIGN-4, info);
     else
 	check_allmem(info);
-  #else
+#else
     if( !p )
 	return;
     if( !(p[-1] == MAGIC_NOR_BYTE || p[-1] == MAGIC_SEC_BYTE) )
 	membug("memory at %p corrupted (underflow=%02x)\n", p, p[-1] );
     else if( p[m_size(p)] != MAGIC_END_BYTE )
 	membug("memory at %p corrupted (overflow=%02x)\n", p, p[-1] );
-  #endif
-  #endif
+#endif
+#endif
 }
 
 
 size_t
 m_size( const void *a )
 {
-  #ifndef M_GUARD
+#ifndef M_GUARD
     log_debug("dummy m_size called\n");
     return 0;
-  #else
+#else
     const byte *p = a;
     size_t n;
 
-  #ifdef M_DEBUG
+#ifdef M_DEBUG
     n = check_mem(p-EXTRA_ALIGN-4, "m_size")->user_n;
-  #else
+#else
     n  = ((byte*)p)[-4];
     n |= ((byte*)p)[-3] << 8;
     n |= ((byte*)p)[-2] << 16;
-  #endif
-    return n;
-  #endif
-}
-
-
-#if 0 /* not used */
-/****************
- * Make a copy of the memory block at a
- */
-void *
-FNAME(copy)( const void *a FNAMEPRT )
-{
-    void *b;
-    size_t n;
-
-    if( !a )
-	return NULL;
-
-    n = m_size(a); Aiiiih woher nehmen
-    if( m_is_secure(a) )
-	b = FNAME(alloc_secure)(n FNAMEARG);
-    else
-	b = FNAME(alloc)(n FNAMEARG);
-    memcpy(b, a, n );
-    return b;
-}
 #endif
+    return n;
+#endif
+}
+
 
 char *
-FNAME(strdup)( const char *a FNAMEPRT )
+FNAMEX(strdup)( const char *a FNAMEPRT )
 {
     size_t n = strlen(a);
-    char *p = FNAME(alloc)(n+1 FNAMEARG);
+    char *p = FNAMEXM(alloc)(n+1 FNAMEARG);
     strcpy(p, a);
     return p;
 }
 
+
+/* Wrapper around xmalloc_clear to take the usual 2 arguments of a
+   calloc style function. */
+void *
+xcalloc (size_t n, size_t m)
+{
+  size_t nbytes;
+
+  nbytes = n * m; 
+  if (m && nbytes / m != n) 
+    out_of_core (nbytes, 0);
+  return xmalloc_clear (nbytes);
+}
+
+/* Wrapper around xmalloc_csecure_lear to take the usual 2 arguments
+   of a calloc style function. */
+void *
+xcalloc_secure (size_t n, size_t m)
+{
+  size_t nbytes;
+
+  nbytes = n * m; 
+  if (m && nbytes / m != n) 
+    out_of_core (nbytes, 1);
+  return xmalloc_secure_clear (nbytes);
+}
+
